from PySide2.QtCore import QAbstractItemModel, QFile, QIODevice, QModelIndex, Qt
from PySide2.QtWidgets import QApplication, QTreeView
from PySide2.QtGui import QImage, QIcon
from pathlib import Path
import os


image_path = os.path.join(Path(os.path.dirname(os.path.realpath(__file__))), 'assets', 'gui_images')
market_icon = QImage(os.path.join(image_path, 'ic_shopping_cart_black_18dp.png'))
production_icon = QImage(os.path.join(image_path, 'ic_build_black_18dp.png'))
import_icon = QImage(os.path.join(image_path, 'ic_keyboard_tab_black_18dp.png'))
cogeneration_icon = QImage(os.path.join(image_path, 'ic_call_split_black_18dp.png'))
unknown_icon = QImage(os.path.join(image_path, 'ic_priority_high_black_18dp.png'))

class TreeItem(object):

    def __init__(self, data, parent=None):

        self.parentItem = parent
        self.itemData = data
        self.childItems = []

    def appendChild(self, item):
        self.childItems.append(item)

    def child(self, row):
        return self.childItems[row]

    def childCount(self):
        return len(self.childItems)

    def columnCount(self):
        return len(self.itemData)

    def data(self, column):
        try:
            return self.itemData#[column]
            #return self.itemData['display']
        except IndexError:
            return None

    def parent(self):
        return self.parentItem

    def row(self):
        if self.parentItem:
            return self.parentItem.childItems.index(self)

        return 0

class TreeModel(QAbstractItemModel):

    def __init__(self, data, parent=None):
        super(TreeModel, self).__init__(parent)

        self.rootItem = TreeItem(['name'])

        self.setupModelData(data, self.rootItem)

    def columnCount(self, parent=None):
        if parent.isValid():
            return parent.internalPointer().columnCount()
        else:
            return self.rootItem.columnCount()

    def data(self, index, role=None):
        if not index.isValid():
            return None

        item = index.internalPointer()

        d = item.data(index.column())

        if role == Qt.DisplayRole:
            return d['display']

        if role == Qt.DecorationRole:
            activity_type = d['type']
            if activity_type == 'market':
                return market_icon
            elif activity_type == 'production':
                return production_icon
            elif activity_type == 'import':
                return import_icon
            elif activity_type == 'cogeneration':
                return cogeneration_icon
            elif activity_type == 'unknown':
                return unknown_icon

        if role == "DataRole":
            return d
        else:
            return None

    def flags(self, index):

        if not index.isValid():
            return Qt.NoItemFlags

        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role=None):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self.rootItem.data(section)
        return None

    def index(self, row, column, parent=None):
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)

        else:
            return QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent()

        if parentItem == self.rootItem:
            return QModelIndex()

        return self.createIndex(parentItem.row(), 0, parentItem)

    def rowCount(self, parent):
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        return parentItem.childCount()

    def get_model_tree(self, data, parent):

        if data:

            treeData = {k: v for k, v in data.items() if k != 'children'}
            treeData['display'] = "[{}] {}".format(data['location'], data['name'])
            #print(treeData['display'])

            thisItem = TreeItem(treeData, parent)

            parent.appendChild(thisItem)

            for c in data['children']:
                self.get_model_tree(c, thisItem)


    def setupModelData(self, data, parent):

        #print(data)

        for sector, tree in data.items():
            this_parent = TreeItem({'name': sector,
                                    'display': sector,
                                    'type': 'toplevel'},
                                   parent)
            parent.appendChild(this_parent)

            for subtree in tree:
                self.get_model_tree(subtree, this_parent)


if __name__ == '__main__':

    import sys
    from futura.constants import ELECTRICITY_TREE

    #print (ELECTRICITY_TREE)

    app = QApplication(sys.argv)

    tree_data = {'Electricity': ELECTRICITY_TREE,
                'Cement':{},
                'Aluminium':{},
                'Steel':{},
                'Lithium':{},
                'Glass':{}
                }

    model = TreeModel(tree_data)

    #print (model)

    view = QTreeView()

    view.setModel(model)
    view.setWindowTitle("TEST")
    view.show()

    sys.exit(app.exec_())








from PySide2 import QtWidgets
from futura_ui.app.ui.utils import load_ui_file
from futura_ui.app.signals import signals
import os


class RunDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super(RunDialog, self).__init__(parent)

        ui_path = 'onrun.ui'
        ui_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), ui_path)

        load_ui_file(ui_path, self)

        self.buttonPressed = None

        self.pushButton_2.pressed.connect(self.pushButton_2_pressed)


    def pushButton_2_pressed(self):
        self.close()
        signals.load_tree_data.emit()





#import wurst as w
from . import w

coal_location_filter = []
coal_location_filter += [w.equals('database', 'ecoinvent')]
coal_location_filter += [w.equals('unit', 'kilowatt hour')]
coal_location_filter += [w.contains('name', 'hard coal')]
coal_location_filter += [w.doesnt_contain_any('name', ['coal mine', 'co-generation'])]

hard_coal_ccs_filter = []
hard_coal_ccs_filter += [w.equals('database', 'Carma CCS')]
hard_coal_ccs_filter += [w.either(*[w.contains('name', 'Hard coal'), w.contains('name', 'hard coal')])]
hard_coal_ccs_filter += [w.contains('location', 'GLO')]

lignite_location_filter = []
lignite_location_filter += [w.equals('database', 'ecoinvent')]
lignite_location_filter += [w.equals('unit', 'kilowatt hour')]
lignite_location_filter += [w.contains('name', 'lignite')]
lignite_location_filter += [w.doesnt_contain_any('name', ['co-generation'])]

lignite_ccs_filter = []
lignite_ccs_filter += [w.equals('database', 'Carma CCS')]
lignite_ccs_filter += [w.either(*[w.contains('name', 'Lignite'), w.contains('name', 'lignite')])]
lignite_ccs_filter += [w.contains('location', 'GLO')]

natural_gas_location_filter = []
natural_gas_location_filter += [w.equals('database', 'ecoinvent')]
natural_gas_location_filter += [w.equals('unit', 'kilowatt hour')]
natural_gas_location_filter += [w.contains('name', 'natural gas')]
natural_gas_location_filter += [w.doesnt_contain_any('name', ['co-generation', 'import', 'aluminium industry', 'burned', '10MW'])]

natural_gas_ccs_filter = []
natural_gas_ccs_filter += [w.equals('database', 'Carma CCS')]
natural_gas_ccs_filter += [w.either(*[w.contains('name', 'Natural gas'), w.contains('name', 'natural gas'), w.contains('name', 'ATR-H2')])]
natural_gas_ccs_filter += [w.doesnt_contain_any('name', ['Synthetic'])]
natural_gas_ccs_filter += [w.contains('location', 'GLO')]

wood_location_filter = []
wood_location_filter += [w.equals('database', 'ecoinvent')]
wood_location_filter += [w.equals('unit', 'kilowatt hour')]
wood_location_filter += [w.contains('name', 'wood')]
wood_location_filter += [w.doesnt_contain_any('name', ['treatment', 'ethanol', 'pellets', 'label-certified', '2000 kW'])]
wood_location_filter += [w.contains('name', 'state-of-the-art')]

wood_ccs_filter = []
wood_ccs_filter += [w.equals('database', 'Carma CCS')]
wood_ccs_filter += [w.either(*[w.contains('name', 'Wood'), w.contains('name', 'wood')])]
wood_ccs_filter += [w.contains('location', 'GLO')]



# -*- coding: utf-8 -*-
import uuid
from io import StringIO

from PySide2 import QtGui
from PySide2.QtUiTools import QUiLoader
from PySide2.QtCore import QMetaObject


class UiLoader(QUiLoader):
    def __init__(self, base_instance):
        QUiLoader.__init__(self, base_instance)
        self.base_instance = base_instance

    def createWidget(self, class_name, parent=None, name=''):
        if parent is None and self.base_instance:
            return self.base_instance
        else:
            # create a new widget for child widgets
            widget = QUiLoader.createWidget(self, class_name, parent, name)
            if self.base_instance:
                setattr(self.base_instance, name, widget)
            return widget


def load_ui_file(ui_file, base_instance=None):
    loader = UiLoader(base_instance)
    widget = loader.load(ui_file)
    QMetaObject.connectSlotsByName(widget)
    return widget


class StdRedirector(StringIO):
    # From http://stackoverflow.com/questions/17132994/pyside-and-python-logging/17145093#17145093
    def __init__(self, widget, out=None, color=None):
        """(edit, out=None, color=None) -> can write stdout, stderr to a
        QTextEdit.
        edit = QTextEdit
        out = alternate stream ( can be the original sys.stdout )
        color = alternate color (i.e. color stderr a different color)
        """
        self.edit_widget = widget
        self.out = out
        self.color = color

    def write(self, text):
        # TODO: Doesn't seem to have any effect
        if self.color:
            original = self.edit_widget.textColor()
            self.edit_widget.setTextColor(QtGui.QColor(self.color))

        self.edit_widget.moveCursor(QtGui.QTextCursor.End)
        self.edit_widget.insertPlainText(text)

        if self.color:
            self.edit_widget.setTextColor(original)

        if self.out:
            self.out.write(text)

    def flush(self, *args, **kwargs):
        pass


def new_id():
    return uuid.uuid4().hex


abt1 = '16z7c78fbfdzb9fbe893c2'



# -*- coding: utf-8 -*-
from .controller import Controller
from .ui.main import MainWindow


class Application(object):
    def __init__(self):
        self.main_window = MainWindow()
        self.controller = Controller()

    def show(self):
        self.main_window.showMaximized()

    def close(self):
        self.main_window.close()

    def deleteLater(self):
        self.main_window.deleteLater()



# -*- coding: utf-8 -*-
import sys
import traceback

from PySide2 import QtWidgets, QtCore

from .application import Application
from .ui.style import default_font


def run_futura():
    qapp = QtWidgets.QApplication(sys.argv)
    # qapp.setFont(default_font)
    application = Application()
    application.show()

    print("Qt Version:", QtCore.__version__)

    def exception_hook(*args):
        print(''.join(traceback.format_exception(*args)))

    sys.excepthook = exception_hook

    sys.exit(qapp.exec_())



#import wurst as w
from . import w
import brightway2 as bw2
from .utils import *
from .storage import storage
import os
from copy import deepcopy
from bw2io.strategies.generic import link_iterable_by_fields

try:
    import _pickle as pickle
except:
    print('falling back on pickle')
    import pickle

import zlib


class FuturaDatabase:

    """
    TODO: Write doctring
    """

    def __init__(self):  # , project_name=None, database_names=None, save_db_file=True):

        self.db = []
        self.database_names = []

        """
        if not database_names:
            database_names = []

        if isinstance(database_names, str):
            database_names = [database_names]

        self.database_names = database_names

        if database_names:
            self.extract_databases(project_name, database_names)

            if save_db_file:
                # TODO: Get rid of this try except statement
                try:
                    self.save()
                except:
                    print('something went wrong')
        """
    def extract_bw2_database(self, project_name, database_names):

        if isinstance(database_names, str):
            database_names = [database_names]
        assert project_name in bw2.projects, "That project doesn't exist"
        assert isinstance(self.database_names, (list, tuple, set)), "Must pass list of database names"

        self.database_names.extend(database_names)

        bw2.projects.set_current(project_name)

        input_db = w.extract_brightway2_databases(database_names)

        fix_unset_technosphere_and_production_exchange_locations(input_db)
        #remove_nones(input_db)

        self.db.extend(input_db)

    def extract_excel_data(self, excelfilepath):
        sp = bw2.ExcelImporter(excelfilepath)

        # link the biosphere exchanges
        sp.apply_strategies(verbose=False)

        # make the internal links
        sp.match_database(fields=["name", "unit", "location"])

        # make the links to existing data
        link_iterable_by_fields(sp.data, self.db, fields=["reference product", "name", "unit", "location"])
        link_iterable_by_fields(sp.data, self.db, fields=["name", "unit", "location"])

        if sp.statistics()[2] != 0:
            fp = sp.write_excel()
            print(fp)
            assert sp.statistics()[2] == 0, "Unlinked exchanges"

        fix_products_and_locations_external(sp.data, self.db)

        self.db.extend(sp.data)

    def write_database(self, project, name, overwrite=False):

        assert project in bw2.projects, "That project doesn't exist"

        if bw2.projects.current != project:
            bw2.projects.set_current(project)

        if name in bw2.databases:
            if not overwrite:
                assert 0, 'Database already exists, either use overwrite=True, or use another name'
            else:
                print('Deleting existing database {}'.format(name))
                del bw2.databases[name]

        w.write_brightway2_database(self.db, name)

    def save(self, save_directory=None, save_filename=None):

        assert self.database_names, 'Nothing to save yet'

        if save_directory is None:
            save_directory = storage.data_dir
            print("Saving to default directory({})".format(save_directory))

        if save_filename is None:
            save_filename = "{}.fdb".format("-".join(self.database_names))
            print("Saving as default filename({})".format(save_filename))

        save_path = os.path.join(save_directory, save_filename)

        with open(save_path, 'wb') as f:
            f.write(zlib.compress(pickle.dumps(self)))

        print('Saved to {}'.format(save_path))

    def load(self, load_path):

        assert self.db == [], "Can only load data into a blank instance of FuturaDatabase"

        with open(load_path, 'rb') as f:
            assert load_path[-4:] == '.fdb', "Not a valid file path"
            print("Loading fdb file from {}".format(load_path))
            loaded = pickle.loads(zlib.decompress(f.read()))

        self.db = loaded.db
        self.database_names = loaded.database_names
        print("Loaded {} with a total of {} activities".format(", ".join(self.database_names), len(self.db)))












# -*- coding: utf-8 -*-
from PySide2 import QtGui, QtWidgets

default_font = QtGui.QFont('Arial', 8)

bold_font = QtGui.QFont()
bold_font.setBold(True)
bold_font.setPointSize(12)

def horizontal_line():
    line = QtWidgets.QFrame()
    line.setFrameShape(QtWidgets.QFrame.HLine)
    line.setFrameShadow(QtWidgets.QFrame.Sunken)
    return line

def vertical_line():
    line = QtWidgets.QFrame()
    line.setFrameShape(QtWidgets.QFrame.VLine)
    line.setFrameShadow(QtWidgets.QFrame.Sunken)
    return line

def header(text):
    label = QtWidgets.QLabel(text)
    label.setFont(bold_font)
    return label


# COLORS values are RGB

class TableStyle:
        # STYLESHEETS
        stylesheet_current_activity = """
        QTableWidget {
            border-radius: 5px;
            background-color: rgb(224, 224, 224);
            border:1px solid rgb(96, 96, 96);
            margin:0px;
            }
        """
        # does not need to have widths for all columns, but starts at col 0
        custom_column_widths = {
            # "ActivitiesTable": [200, 250, 50],
        }

class ActivitiesTab:
    style_sheet_read_only ="""
        QTabWidget::pane {
            border-top: 0px solid rgb(128,0,0); /*red line (read-only indicator) - removed due to request */
            /*border-bottom: 3px solid rgb(128,0,0);*/
        }        
    """
    style_sheet_editable = """
        QTabWidget::pane {
            border-top: 3px solid rgb(0,128,0);
            /* border-bottom: 3px solid rgb(0,128,0);*/
        }        
        """


class ActivitiesPanel:
    style_sheet = """
    """

class TableItemStyle:
    COLOR_CODE = {
        'default': (0, 0, 0),  # black
        # 'product': (0, 132, 130),
        'product': (0, 0, 0),
        # 'reference product': (0, 132, 130),
        'reference product': (0, 0, 0),
        'name': (0, 2, 140),
        'activity': (0, 72, 216),
        'amount': (0, 0, 0),
        # 'unit': (51, 153, 255),
        'unit': (0, 0, 0),
        'location': (72, 0, 140),
        'database': (96, 96, 96),
        'categories': (0, 0, 0),
        'key': (96, 96, 96),
    }

    def __init__(self):
        self.brushes = {}
        for key, values in self.COLOR_CODE.items():
            self.brushes.update({
                key: QtGui.QBrush(QtGui.QColor(*values))
            })


style_activity_panel = ActivitiesPanel
style_activity_tab = ActivitiesTab
style_table = TableStyle()
style_item = TableItemStyle()



#import wurst as w
from . import w

from futura.wrappers import FuturaDatabase

import warnings

def create_regional_activities(base_activity, new_regions, db, production_volumes=None,
                               remove_production_from_original=True, relink_now = True):
    if production_volumes:
        assert len(production_volumes) == len(new_regions)
        total_production = 0

    geomatcher = w.geomatcher
    added_datasets = []
    code_list = []
    for n, region in enumerate(new_regions):
        if region not in geomatcher.contained(base_activity['location']):
            # print("{} not found within {}".format(region, base_activity['location']))
            #warnings.warn("{} not found within {}".format(region, base_activity['location']))
            pass
        new_ds = w.copy_to_new_location(base_activity, region)

        if 'input' in w.reference_product(new_ds).keys():
            # print("Deleting input from production exchange {} (from {})".format(
            # w.reference_product(new_ds)['name'], new_ds['name']))
            del w.reference_product(new_ds)['input']

        for e in w.technosphere(new_ds):
            if 'input' in e.keys():
                # print("Deleting input from {} (input to {})".format(e['name'], new_ds['name']))
                del e['input']

        if relink_now:
            new_ds = w.relink_technosphere_exchanges(new_ds,
                                                     db,
                                                     exclusive=True,
                                                     drop_invalid=False,
                                                     biggest_first=False,
                                                     contained=False,
                                                     exclude=['UCTE'])
        else:
            code_list.append((new_ds['database'], new_ds['code']))

        if production_volumes:
            pv = production_volumes[n]
            production_exchanges = [e for e in w.production(new_ds)]
            assert len(production_exchanges) == 1, "This process has multiple or no outputs - needs to have exactly 1"
            production_exchange = production_exchanges[0]

            production_exchange['production volume'] = pv

            total_production += pv

        added_datasets.append("{} [{}]".format(new_ds['name'], new_ds['location']))
        db.append(new_ds)

    if production_volumes:
        if remove_production_from_original:
            production_exchanges = [e for e in w.production(base_activity)]
            assert len(production_exchanges) == 1, "This process has multiple or no outputs - needs to have exactly 1"
            production_exchange = production_exchanges[0]
            original_pv = production_exchange['production volume']
            production_exchange['production volume'] -= total_production
            print("Changed production volume for {} [{}] from {} to {}".format(base_activity['name'],
                                                                               base_activity['location'], original_pv,
                                                                               production_exchange[
                                                                                   'production volume']))

    #print("Added the following datasets\n{}".format(added_datasets))
    if not relink_now:
        return code_list



from wurst import log
from wurst.errors import InvalidLink
from wurst.geo import geomatcher
from wurst.searching import reference_product, get_many, equals, get_one
from wurst.transformations.uncertainty import rescale_exchange
from wurst.transformations.utils import copy_dataset
from constructive_geometries import resolved_row
from copy import deepcopy


def copy_to_new_location(ds, location):
    """Copy dataset and substitute new ``location``.

    Doesn't change exchange locations, except for production exchanges.

    Returns the new dataset."""
    MESSAGE = "Copied activity from '{old}' location to '{new}'."
    log({
        'function': 'copy_to_new_location',
        'message': MESSAGE.format(old=ds['location'], new=location)
    }, ds)

    cp = copy_dataset(ds)
    cp['location'] = location
    for exc in cp['exchanges']:
        if exc['type'] == 'production':
            exc['location'] = location

    return cp


def relink_technosphere_exchanges(ds, data, exclusive=True,
                                  drop_invalid=False, biggest_first=False, contained=True, exclude=None):
    """Find new technosphere providers based on the location of the dataset.

    Designed to be used when the dataset's location changes, or when new datasets are added.

    Uses the name, reference product, and unit of the exchange to filter possible inputs. These must match exactly. Searches in the list of datasets ``data``.

    Will only search for providers contained within the location of ``ds``, unless ``contained`` is set to ``False``, all providers whose location intersects the location of ``ds`` will be used.

    A ``RoW`` provider will be added if there is a single topological face in the location of ``ds`` which isn't covered by the location of any providing activity.

    If no providers can be found, `relink_technosphere_exchanes` will try to add a `RoW` or `GLO` providers, in that order, if available. If there are still no valid providers, a ``InvalidLink`` exception is raised, unless ``drop_invalid`` is ``True``, in which case the exchange will be deleted.

    Allocation between providers is done using ``allocate_inputs``; results seem strange if ``contained=False``, as production volumes for large regions would be used as allocation factors.

    Input arguments:

        * ``ds``: The dataset whose technosphere exchanges will be modified.
        * ``data``: The list of datasets to search for technosphere product providers.
        * ``exclusive``: Bool, default is ``True``. Don't allow overlapping locations in input providers.
        * ``drop_invalid``: Bool, default is ``False``. Delete exchanges for which no valid provider is available.
        * ``biggest_first``: Bool, default is ``False``. Determines search order when selecting provider locations. Only relevant is ``exclusive`` is ``True``.
        * ``contained``: Bool, default is ``True``. If ture, only use providers whose location is completely within the ``ds`` location; otherwise use all intersecting locations.
        * ``exclude``: TODO: write this

    Modifies the dataset in place; returns the modified dataset."""

    #print("MonkeyPatch!! relink_technosphere_exchanges")

    MESSAGE = "Relinked technosphere exchange of {}/{}/{} from {}/{} to {}/{}."
    DROPPED = "Dropped technosphere exchange of {}/{}/{}; no valid providers."
    new_exchanges = []
    technosphere = lambda x: x['type'] == 'technosphere'

    for exc in filter(technosphere, ds['exchanges']):
        possible_datasets = list(get_possibles(exc, data))
        possible_locations = [obj['location'] for obj in possible_datasets]
        if exclude:
            possible_locations = [x for x in possible_locations if x not in exclude]
        with resolved_row(possible_locations, geomatcher) as g:
            func = g.contained if contained else g.intersects
            gis_match = func(ds['location'], include_self=True, exclusive=exclusive,
                             biggest_first=biggest_first, only=possible_locations)

        kept = [ds for loc in gis_match for ds in possible_datasets
                if ds['location'] == loc]

        if kept:
            with resolved_row(possible_locations, geomatcher) as g:
                missing_faces = geomatcher[ds['location']].difference(
                    set.union(*[geomatcher[obj['location']] for obj in kept])
                )
            if missing_faces and "RoW" in possible_locations:
                kept.extend([obj for obj in possible_datasets if obj['location'] == 'RoW'])
        elif 'RoW' in possible_locations:
            kept = [obj for obj in possible_datasets if obj['location'] == 'RoW']

        if not kept and "GLO" in possible_locations:
            kept = [obj for obj in possible_datasets if obj['location'] == 'GLO']

        if not kept:
            if drop_invalid:
                log({
                    'function': 'relink_technosphere_exchanges',
                    'message': DROPPED.format(
                        exc['name'], exc['product'], exc['unit']
                    )
                }, ds)
                continue
            else:
                print("technosphere exchange of {}/{}/{}; no valid providers.".format(exc['name'], exc['product'],
                                                                                      exc['unit']))
                raise InvalidLink

        allocated = allocate_inputs(exc, kept)

        for obj in allocated:
            log({
                'function': 'relink_technosphere_exchanges',
                'message': MESSAGE.format(
                    exc['name'], exc['product'], exc['unit'], exc['amount'],
                    ds['location'], obj['amount'], obj['location']
                )
            }, ds)

        new_exchanges.extend(allocated)

    ds['exchanges'] = [
                          exc for exc in ds['exchanges']
                          if exc['type'] != 'technosphere'
                      ] + new_exchanges
    return ds


def allocate_inputs(exc, lst):
    """Allocate the input exchanges in ``lst`` to ``exc``, using production volumes where possible, and equal splitting otherwise.

    Always uses equal splitting if ``RoW`` is present."""

    #print("MonkeyPatch!! allocate_inputs")

    MESSAGE = "Changed technosphere exchange of {}/{} to {}/{}."

    has_row = any((x['location'] in ('RoW', 'GLO') for x in lst))
    pvs = [reference_product(o).get('production volume') or 0 for o in lst]
    if all((x > 0 for x in pvs)) and not has_row:
        # Allocate using production volume
        total = sum(pvs)
    else:
        # Allocate evenly
        total = len(lst)
        pvs = [1 for _ in range(total)]

    def new_exchange(exc, obj, factor):
        cp = deepcopy(exc)

        if cp['name'] != obj['name']:
            log({
                'function': 'allocate_inputs',
                'message': MESSAGE.format(
                    cp['name'], cp['location'],
                    obj['name'], obj['location']
                )
            }, cp)

            cp['name'] = obj['name']

        cp['location'] = obj['location']

        return rescale_exchange(cp, factor)

    return [
        new_exchange(exc, obj, factor / total)
        for obj, factor in zip(lst, pvs)
    ]


def get_possibles(exchange, data):
    """Filter a list of datasets ``data``, returning those with the same name, reference product, and unit as in ``exchange``.

    Returns a generator."""

    market_name = exchange['name']
    if market_name.startswith('market group'):
        market_name = 'market' + market_name[len('market group'):]

    if market_name != exchange['name']:
        key = (market_name, exchange['product'], exchange['unit'])
        for ds in data:
            if (ds['name'], ds['reference product'], ds['unit']) == key:
                yield ds
    try:
        key = (exchange['name'], exchange['product'], exchange['unit'])
    except KeyError:
        print(exchange)
        assert 0
    for ds in data:
        if (ds['name'], ds['reference product'], ds['unit']) == key:
            yield ds


def default_global_location(database):
    """Set missing locations to ```GLO``` for datasets in ``database``.

    Changes location if ``location`` is missing or ``None``. Will add key ``location`` if missing."""
    for ds in get_many(database, *[equals('location', None)]):
        ds['location'] = 'GLO'
    return database



#import wurst as w
from . import w

import pprint
from wurst.searching import exclude


def fix_unset_technosphere_and_production_exchange_locations(db, matching_fields=('name', 'unit')):

    """
    Utility function from wurst publication supplementary materials to fix unset technosphere and production
    exchanges.
    Database is fixed in place, function returns nothing

    :param db: database to fix
    :param matching_fields: fields on which to search for exchanges
    :return: nothing
    """
    for ds in db:
        for exc in ds['exchanges']:
            if exc['type'] == 'production' and exc.get('location') is None:
                exc['location'] = ds['location']
            elif exc['type'] == 'technosphere' and exc.get('location') is None:
                locs = find_location_given_lookup_dict(db,
                                                       {k: exc.get(k) for k in matching_fields})
                if len(locs) == 1:
                    exc['location'] = locs[0]
                else:
                    print("No unique location found for exchange:\n{}\nFound: {}".format(
                        pprint.pformat(exc), locs
                    ))


def find_location_given_lookup_dict(db, lookup_dict):
    """
    Utility function for the utility function above
    :param db: database to fix
    :param lookup_dict: dictionary of locations
    :return: list of locations
    """
    return [x['location'] for x in w.get_many(db, *[w.equals(k, v) for k, v in lookup_dict.items()])]


def remove_nones(db):
    exists = lambda x: {k: v for k, v in x.items() if v is not None}

    for ds in db:

        ds['exchanges'] = [exists(exc) for exc in ds['exchanges']]


def fix_products_and_locations_external(external_data, existing_data):
    external_db_set = set([a['database'] for a in external_data])
    internal_db_set = set([a['database'] for a in existing_data])
    # print(external_db_set)
    # print(internal_db_set)

    # assert 0
    product_exchange_count = 0
    location_exchange_count = 0
    location_activity_count = 0
    production_volume_count = 0

    for a in external_data:

        # add missing location to activity
        if not 'location' in a.keys():
            a['location'] = 'GLO'
            location_activity_count += 1
            # print('Added GLO location to {}'.format(a['name']))

        for e in [x for x in a['exchanges'] if x['type'] != 'biosphere']:

            # add missing location

            if e['input'][0] in external_db_set:
                look_in = external_data
            elif e['input'][0] in internal_db_set:
                look_in = existing_data
            else:
                # print(e['input'][0])
                continue

            this_input = e['input']
            # print(this_input)

            input_filter = [w.equals('database', this_input[0])]
            input_filter += [w.equals('code', this_input[1])]

            try:
                this_input = w.get_one(look_in, *input_filter)
            except w.errors.NoResults:
                assert 0, "{} not found".format(this_input)

            if not 'product' in e.keys():
                e['product'] = this_input['reference product']
                product_exchange_count += 1

            if not 'location' in e.keys():
                if 'location' in this_input.keys():
                    e['location'] = this_input['location']
                    # print('Added {} location to exchange {} of {}'.format(this_input['location'], e['name'], a['name']))
                else:
                    e['location'] = 'GLO'
                    # print('Added GLO location to exchange {} of {}'.format(e['name'], a['name']))

                location_exchange_count += 1

            # add missing production volumes (default to zero)

            if e['type'] == 'production':
                if 'production volume' not in e.keys():
                    e['production volume'] = 0
                    production_volume_count += 1
                    # print('Added default zero production volume to exchange of {} from {}'.format(e['product'], a['name']))

    print("Location data added to {} activities".format(location_activity_count))
    print("Location data added to {} exchanges".format(location_exchange_count))
    print("Product data added to {} exchanges".format(product_exchange_count))
    print("Default (zero) production volume data added to {} production exchanges".format(production_volume_count))


def create_filter_from_description(description):

    wurst_functions = {
        'equals': w.equals,
        'contains': w.contains,
        'startswith': w.startswith,
        'either': w.either,
        'exclude': exclude,
        'doesnt_contain_any': w.doesnt_contain_any,
    }

    def create_filter(detail):

        if detail['filter'] in ['either', 'exclude']:
            sub_filter = []

            for x in detail['args']:
                sub_filter += create_filter(x)

            return [wurst_functions[detail['filter']](*sub_filter)]

        return [wurst_functions[detail['filter']](*detail['args'])]

    this_filter = []

    for x in description:
        this_filter += create_filter(x)

    return this_filter



# -*- coding: utf-8 -*-
import sys

from PySide2 import QtCore, QtGui, QtWidgets
#from PySide2.QtUiTools import QUiLoader
from PySide2.QtCore import QFile

from .utils import load_ui_file

# from .style import header
from .icons import qicons
# from .menu_bar import MenuBar
# from .panels import LeftPanel, RightPanel
# from .statusbar import Statusbar
# from .utils import StdRedirector
from ..signals import signals
import os

APP_NAME = "Futura"
APP_VERSION = "0.0.1"


class MainWindow(QtWidgets.QMainWindow):
    # DEFAULT_NO_METHOD = 'No method selected yet'

    def __init__(self):
        super(MainWindow, self).__init__(None)

        ui_path = os.path.join('main', 'main.ui')
        ui_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), ui_path)

        load_ui_file(ui_path, self)
        #self.ui.setupUi(self)

        self.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))

        # Window title
        self.setWindowTitle("{} {}".format(APP_NAME, APP_VERSION))

        # Background Color
        # self.setAutoFillBackground(True)
        # p = self.palette()
        # p.setColor(self.backgroundRole(), QtGui.QColor(148, 143, 143, 127))
        # self.setPalette(p)

        # Small icon in main window titlebar
        self.icon = qicons.futura
        self.setWindowIcon(self.icon)
        t = QtCore.QTimer()
        t.singleShot(0, self.show_rundialog)

    def show_rundialog(self):
        signals.run_dialog.emit(self)


        # self.connect_signals()

    # def connect_signals(self):
    #     # Keyboard shortcuts
    #     self.shortcut_debug = QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+D"), self)
    #     self.shortcut_debug.activated.connect(self.toggle_debug_window)
    #
    # def toggle_debug_window(self):
    #     """Toggle between any window and the debug window."""
    #     if self.stacked.currentWidget() != self.debug_widget:
    #         self.last_widget = self.stacked.currentWidget()
    #         self.stacked.setCurrentWidget(self.debug_widget)
    #         # print("Switching to debug window")
    #     else:
    #         # print("Switching back to last widget")
    #         if self.last_widget:
    #             try:
    #                 self.stacked.setCurrentWidget(self.last_widget)
    #             except:
    #                 print("Previous Widget has been deleted in the meantime. Switching to main window.")
    #                 self.stacked.setCurrentWidget(self.main_widget)
    #         else:  # switch to main window
    #             self.stacked.setCurrentWidget(self.main_widget)
    #
    # def add_tab_to_panel(self, obj, label, side):
    #     panel = self.left_panel if side == 'left' else self.right_panel
    #     panel.addTab(obj, label)
    #
    # def select_tab(self, obj, side):
    #     panel = self.left_panel if side == 'left' else self.right_panel
    #     panel.setCurrentIndex(panel.indexOf(obj))
    #
    # def dialog(self, title, label):
    #     value, ok = QtWidgets.QInputDialog.getText(self, title, label)
    #     if ok:
    #         return value
    #
    # def info(self, label):
    #     QtWidgets.QMessageBox.information(
    #         self,
    #         "Information",
    #         label,
    #         QtWidgets.QMessageBox.Ok,
    #     )
    #
    # def warning(self, title, text):
    #     QtWidgets.QMessageBox.warning(self, title, text)
    #
    # def confirm(self, label):
    #     response = QtWidgets.QMessageBox.question(
    #         self,
    #         "Confirm Action",
    #         label,
    #         QtWidgets.QMessageBox.Yes,
    #         QtWidgets.QMessageBox.No
    #     )
    #     return response == QtWidgets.QMessageBox.Yes



from . import warn
from .wrappers import FuturaDatabase
#import wurst as w
from . import w
from wurst.searching import exclude
from itertools import groupby
from collections import OrderedDict
import pandas as pd
import matplotlib.pyplot as plt

def add_exchange_to_activity(base_activity, activity_to_link_to):
    exchange_template = {'uncertainty type': 0,
                         'loc': 0,
                         'amount': 0,
                         'type': 'technosphere',
                         'production volume': 0,
                         'product': '',
                         'name': '',
                         'unit': '',
                         'location': ''}

    rp = w.reference_product(activity_to_link_to)
    key_list = ['production volume', 'product', 'name', 'unit', 'location']
    for k in key_list:
        exchange_template[k] = rp[k]

    if exchange_template not in base_activity['exchanges']:
        base_activity['exchanges'].append(exchange_template)
    else:
        warn('Exchange of {} already exists in this process, not adding it'.format(activity_to_link_to['name']))


def get_processes_from_exchanges(process, database, reference_product):

    assert isinstance(database, FuturaDatabase), "database needs to be a futura FuturaDatabase object"

    processes = [w.get_one(database.db, *[w.equals('name', e['name']),
                                          w.equals('location', e['location']),
                                          w.equals('reference product', e['product'])])
                 for e in w.technosphere(process)
                 if e['product'] == reference_product and e['name'] != process['name']]

    return processes


def alter_production_volumes(processes, production_volumes, mismatch_ok=False):

    old_pvs = [w.reference_product(x)['production volume'] for x in processes]
    pre_total = sum(old_pvs)

    assert len(processes) == len(production_volumes)

    if not mismatch_ok:
        assert pre_total == sum(
            production_volumes), "The old and new production volumes don't match, " \
                                 "either fix this or set mismatch_ok = True"

    for n, x in enumerate(processes):
        w.reference_product(x)['production volume'] = production_volumes[n]


def fix_exchange_production_volumes(process, database):

    assert isinstance(database, FuturaDatabase), "database needs to be a futura FuturaDatabase object"

    for e in w.technosphere(process):
        link_filter = [w.equals('name', e['name']), w.equals('location', e['location']),
                       w.equals('reference product', e['product'])]
        actual_link = w.get_one(database.db, *link_filter)
        e['production volume'] = w.reference_product(actual_link)['production volume']


def update_technosphere_exchanges_from_pvs(process, database):

    assert isinstance(database, FuturaDatabase), "database needs to be a futura FuturaDatabase object"

    fix_exchange_production_volumes(process, database)

    technosphere_exchanges = [e for e in w.technosphere(process)]

    for g, v in groupby(technosphere_exchanges, lambda x: x['product']):
        # print(g)
        v_list = list(v)
        self_excluded_v_list = [x for x in v_list if x['name'] != process['name']]

        amount_sum = sum([x['amount'] for x in self_excluded_v_list])

        pv_sum = sum(x['production volume'] for x in self_excluded_v_list)

        for x in self_excluded_v_list:
            new_amount = amount_sum * (x['production volume'] / pv_sum)
            # print("{}: {}".format(x['name'], new_amount))
            x['amount'] = new_amount
            x['loc'] = new_amount


def find_possible_additional_market_exchanges(process, database, include_transport=False):
    assert isinstance(database, FuturaDatabase), "database needs to be a futura FuturaDatabase object"
    assert process['name'].startswith('market') and not process['name'].startswith('market group'), \
        'This function only works with market processes (not market groups or production processes)'

    technosphere_exchanges = [e for e in w.technosphere(process)]

    result = []
    # result_dict = {}

    for g, v in groupby(technosphere_exchanges, lambda x: (x['product'],
                                                           x['location'],
                                                           x['unit'])):
        v_list = list(v)
        names = [e['name'] for e in v_list if e['name'] != process['name']]

        possible_additions_filter = []
        possible_additions_filter += [w.equals('unit', g[2])]
        possible_additions_filter += [w.equals('reference product', g[0])]
        possible_additions_filter += [exclude(w.startswith('name', 'market'))]
        possible_additions_filter += [w.doesnt_contain_any('name', ['production mix'])]
        possible_additions_filter += [w.equals('location', g[1])]
        possible_additions_filter += [w.doesnt_contain_any('name', names)]

        if not include_transport:
            possible_additions_filter += [w.doesnt_contain_any('name', ['transport'])]

        possibles = list(w.get_many(database.db, *possible_additions_filter))

        if len(possibles) > 0:
            result.extend(possibles)
            # result_dict[g] = possibles

    return result  # , result_dict


def get_input_processes_to_market(process, database):
    reference_product = w.reference_product(process)['product']

    return get_processes_from_exchanges(process, database, reference_product)


class FuturaMarket:
    def __init__(self, market, database):
        self.market = market
        self.database = database
        self.processes = get_input_processes_to_market(market, database)
        self.process_dict = {x['name']: {'process': x, 'production volume': w.reference_product(x)['production volume']}
                             for x in self.processes}

    def __repr__(self):
        return "FuturaMarket for {}\n\n{}".format(self.market['name'], str(dict(self.sorted_percentages)))

    @property
    def production_volumes(self):
        return [x['production volume'] for k, x in self.process_dict.items()]

    @property
    def total_production(self):
        return sum(self.production_volumes)

    @property
    def percentages(self):
        return {k: round(v['production volume'] / self.total_production, 4) for k, v in self.process_dict.items()}

    @property
    def sorted_percentages(self):
        sorted_x = sorted(self.percentages.items(), key=lambda kv: kv[1], reverse=True)
        sp = OrderedDict(sorted_x)
        df = pd.DataFrame(sorted_x, columns=['item', 'percentage'])
        df = df.set_index('item')
        # df = df.style.format({'percentage': '{0:.2%}'})
        return df

    @property
    def plot(self):
        return self.sorted_percentages.plot(kind='pie', y='percentage', legend=None)

    def get_pv(self, process_name):
        return self.process_dict[process_name]['production volume']

    def set_pv(self, process_name, new_pv):
        self.process_dict[process_name]['production volume'] = new_pv

    def add_pv(self, process_name, new_pv):
        self.process_dict[process_name]['production volume'] += new_pv

    def subtract_pv(self, process_name, new_pv):
        self.process_dict[process_name]['production volume'] -= new_pv

    def transfer_pv(self, from_name, to_name, factor=None, amount=None):

        assert any([factor, amount]), "you need to set either a factor or an amount"
        assert not all([factor, amount]), "you can't set a factor and an amount, choose one"

        from_pv = self.get_pv(from_name)

        if amount:
            assert amount <= from_pv, "you can't transfer more pv than is available"
            self.subtract_pv(from_name, amount)
            self.add_pv(to_name, amount)

        elif factor:
            assert 0 <= factor <= 1, "factor must be between 0 and 1"
            transfer_amount = from_pv*factor
            self.subtract_pv(from_name, transfer_amount)
            self.add_pv(to_name, transfer_amount)

    def rewrite_pvs(self):
        for name, process in self.process_dict.items():
            if w.reference_product(process['process'])['production volume'] != process['production volume']:
                w.reference_product(process['process'])['production volume'] = process['production volume']

    def relink(self):
        self.rewrite_pvs()
        update_technosphere_exchanges_from_pvs(self.market, self.database)

    def add_alternative_exchanges(self, include_transport=False):
        possibles = find_possible_additional_market_exchanges(self.market, self.database, include_transport)
        for p in possibles:
            add_exchange_to_activity(self.market, p)

        self.processes = get_input_processes_to_market(self.market, self.database)

        for x in self.processes:
            if x['name'] not in self.process_dict.keys():
                self.process_dict[x['name']] = {'process': x,
                                                'production volume': w.reference_product(x)['production volume']}




# -*- coding: utf-8 -*-
from PySide2.QtCore import QObject, Signal


class Signals(QObject):
    """ Signals used for the Activity Browser should be defined here.
    While arguments can be passed to signals, it is good practice not to do this if possible. """
    run_dialog = Signal(object)
    load_tree_data = Signal()
    # # General Settings
    #
    # # bw2 directory
    # switch_bw2_dir_path = Signal(str)
    # # directory_changed = Signal()
    #
    # # Project
    change_project = Signal(str)
    # # change_project_dialog = Signal()
    # new_project = Signal()
    # copy_project = Signal()
    # delete_project = Signal()
    project_selected = Signal()
    # projects_changed = Signal()
    #
    # # Database
    # add_database = Signal()
    # delete_database = Signal(str)
    # copy_database = Signal(str)
    # install_default_data = Signal()
    # import_database = Signal()
    #
    # database_selected = Signal(str)
    # databases_changed = Signal()
    # database_changed = Signal(str)
    # database_read_only_changed = Signal(str, bool)
    #
    # # Activity (key, field, new value)
    # new_activity = Signal(str)
    # add_activity_to_history = Signal(tuple)
    # duplicate_activity = Signal(tuple)
    # duplicate_activity_to_db = Signal(str, object)
    # show_duplicate_to_db_interface = Signal(tuple)
    # open_activity_tab = Signal(tuple)
    # open_activity_graph_tab = Signal(tuple)
    # delete_activity = Signal(tuple)
    #
    # # Activity editing
    # edit_activity = Signal(str)  # db_name
    # activity_modified = Signal(tuple, str, object)
    #
    # # Exchanges
    # # exchanges_output_modified = Signal(list, tuple)
    # exchanges_deleted = Signal(list)
    # exchanges_add = Signal(list, tuple)
    # exchange_amount_modified = Signal(object, float)
    # exchange_modified = Signal(object, str, object)
    #
    # # Parameters
    # add_activity_parameter = Signal(tuple)
    # parameters_changed = Signal()
    # # Pass the key of the activity holding the exchange
    # exchange_formula_changed = Signal(tuple)
    #
    # # Calculation Setups
    # new_calculation_setup = Signal()
    # delete_calculation_setup = Signal(str)
    # rename_calculation_setup = Signal(str)
    # set_default_calculation_setup = Signal()
    #
    # calculation_setup_changed = Signal()
    # calculation_setup_selected = Signal(str)
    #
    # # LCA Results
    # lca_calculation = Signal(str)
    # lca_results_tabs_changed = Signal()
    #
    # method_selected = Signal(tuple)
    # method_tabs_changed = Signal()
    #
    # # Qt Windows
    # update_windows = Signal()
    # new_statusbar_message = Signal(str)
    #
    # # Tabs
    # toggle_show_or_hide_tab = Signal(str)
    # show_tab = Signal(str)
    # hide_tab = Signal(str)
    # hide_when_empty = Signal()
    #
    # # Metadata
    # metadata_changed = Signal(tuple)  # key


signals = Signals()



from PySide2 import QtWidgets
from futura_ui.app.ui.utils import load_ui_file
from futura_ui.app.ui.tables.projects import ProjectListWidget
from futura_ui.app.signals import signals
import os


class ProjectDialog(QtWidgets.QDialog):
    def __init__(self, parent=None):
        super(ProjectDialog, self).__init__(parent)

        ui_path = 'project_choice.ui'
        ui_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), ui_path)

        load_ui_file(ui_path, self)

        projectList = ProjectListWidget()

        self.verticalLayout.addWidget(projectList)




import brightway2 as bw
#import wurst as w
from . import w

electricity_markets = [w.contains('unit', 'kilowatt hour')]
electricity_markets += [w.contains('name', 'electricity')]

cement_filter = [w.contains('name', 'cement')]
#cement_filter += [w.doesnt_contain_any('name','factory')]

def build_tree(activity, max_levels=2, amount=None, filters=[], seen_list=[], level=0):
    if amount is None:
        amount = 1

    if isinstance(activity, tuple):
        activity = bw.get_activity(activity)

    inputs = list(activity.technosphere())
    # print(list(inputs[0].keys()))
    # print(inputs[0]['flow'])

    include = list(w.get_many(inputs, *filters))
    include = [exc for exc in include if exc['input'][0] == activity['database']]
    duplicate = activity['code'] in seen_list

    if activity['name'].startswith('market'):
        activity_type = 'market'
    elif 'production' in activity['name']:
        activity_type = 'production'
    elif 'import' in activity['name']:
        activity_type = 'import'
    elif 'co-generation' in activity['name']:
        activity_type = 'cogeneration'
    else:
        activity_type = 'unknown'

    # inside = [exc for exc in inputs
    #          if exc['input'][0] == activity['database'] and exc['unit'] == 'kilowatt hour']

    if level <= max_levels and duplicate == False:
        technosphere = [
            build_tree(exc.input, max_levels, exc['amount'], filters, seen_list=seen_list + [activity['code']],
                       level=level + 1)
            for exc in include]
    else:
        technosphere = []

    return {
        'name': activity['name'],
        'location': activity['location'],
        'amount': amount,
        'children': technosphere,
        'id': activity['code'],
        'duplicate': duplicate,
        'type': activity_type,
    }


# -*- coding: utf-8 -*-

#from activity_browser.app.ui.wizards.db_import_wizard import (
#    DatabaseImportWizard, DefaultBiosphereDialog, CopyDatabaseDialog
#)
#from .bwutils import commontasks as bc, AB_metadata

from futura_ui.app.ui.dialogs.run_dialog import RunDialog
from futura_ui.app.ui.dialogs.project_choice import ProjectDialog

from .settings import futura_settings # , project_settings
from .signals import signals

import brightway2 as bw

class Controller(object):
    """The Controller is a central object in the Activity Browser. It groups methods that may be required in different
    parts of the AB to access, modify, or delete:
    - settings
    - projects
    - databases
    - calculation setups
    - activities/exchanges
    It is different from bwutils in that it also contains Qt elements such as dialogs.
    - """
    def __init__(self):
        self.connect_signals()              # connect the signals the app requires - this is what the app does
        # signals.project_selected.emit()     # emit the project_selected signal - this runs the self.ensure_sqlite_indices function
        # signals.run_dialog.emit(None)           # emit the run_dialog signal
        self.load_settings()                # load settings
        self.db_wizard = None               # create an attribute for the db_wizard
        #print('Brightway2 data directory: {}'.format(bw.projects._base_data_dir))
        #print('Brightway2 active project: {}'.format(bw.projects.current))

    def connect_signals(self):
        signals.run_dialog.connect(self.run_dialog)
        signals.load_tree_data.connect(self.load_tree_data)

        # Note - these need to be set up in the signals file too

        ## SLOTS
        ### Project
        signals.project_selected.connect(self.project_selected)
        # signals.new_project.connect(self.new_project)
        signals.change_project.connect(self.change_project)
        # signals.change_project_dialog.connect(self.change_project_dialog)
        # signals.copy_project.connect(self.copy_project)
        # signals.delete_project.connect(self.delete_project)

        ### Database
        # signals.add_database.connect(self.add_database)
        # signals.delete_database.connect(self.delete_database)
        # signals.copy_database.connect(self.copy_database)
        # signals.install_default_data.connect(self.install_default_data)
        # signals.import_database.connect(self.import_database_wizard)

        ### Activity
        # signals.duplicate_activity.connect(self.duplicate_activity)
        # signals.activity_modified.connect(self.modify_activity)
        # signals.new_activity.connect(self.new_activity)
        # signals.delete_activity.connect(self.delete_activity)
        # signals.duplicate_activity_to_db.connect(self.duplicate_activity_to_db)
        # signals.show_duplicate_to_db_interface.connect(self.show_duplicate_to_db_interface)

        ### Exchange
        # signals.exchanges_output_modified.connect(self.modify_exchanges_output)
        # signals.exchanges_deleted.connect(self.delete_exchanges)
        # signals.exchanges_add.connect(self.add_exchanges)
        # signals.exchange_amount_modified.connect(self.modify_exchange_amount)
        # signals.exchange_modified.connect(self.modify_exchange)

        ### Calculation Setups
        # signals.new_calculation_setup.connect(self.new_calculation_setup)
        # signals.rename_calculation_setup.connect(self.rename_calculation_setup)
        # signals.delete_calculation_setup.connect(self.delete_calculation_setup)

        ### Other
        # signals.switch_bw2_dir_path.connect(self.switch_brightway2_dir_path)

# SETTINGS
    def load_settings(self):
        if futura_settings.settings:
            print("Loading user settings:")
            # self.switch_brightway2_dir_path(dirpath=futura_settings.custom_bw_dir)
            # self.change_project(futura_settings.startup_project)

# Specific functions to do things go here, within the controller class

    #def import_database_wizard(self):
    #    try:
    #        if self.db_wizard is None:
    #            self.db_wizard = DatabaseImportWizard()
    #        else:
    #            self.db_wizard.show()
    #            self.db_wizard.activateWindow()
    #    except:
    #        QtWidgets.QMessageBox.warning(None,
    #                                          "Error during importing.",
    #                                          "Oops. Something went wrong with the data import. "
    #                                          "Please check the console for details.")

    #def switch_brightway2_dir_path(self, dirpath):
    #    if dirpath == bw.projects._base_data_dir:
    #        return  # dirpath is already loaded
    #    try:
    #        assert os.path.isdir(dirpath)
    #        bw.projects._base_data_dir = dirpath
    #        bw.projects._base_logs_dir = os.path.join(dirpath, "logs")
    #        # create folder if it does not yet exist
    #        if not os.path.isdir(bw.projects._base_logs_dir):
    #            os.mkdir(bw.projects._base_logs_dir)
    #        # load new brightway directory
    #        bw.projects.db = SubstitutableDatabase(
    #            os.path.join(bw.projects._base_data_dir, "projects.db"),
    #            [ProjectDataset]
    #        )
    #        print('Loaded brightway2 data directory: {}'.format(bw.projects._base_data_dir))
    #        self.change_project(futura_settings.startup_project, reload=True)
    #        signals.databases_changed.emit()

    #    except AssertionError:
    #        print('Could not access BW_DIR as specified in settings.py')

# PROJECT
#     def change_project_dialog(self):
#         project_names = sorted([x.name for x in bw.projects])
#         name, ok = QtWidgets.QInputDialog.getItem(
#             None,
#             "Choose project",
#             "Name:",
#             project_names,
#             project_names.index(bw.projects.current),
#             False
#         )
#         if ok:
#             self.change_project(name)
#
    def change_project(self, name=None, reload=False):
        # TODO: what should happen if a new project is opened? (all activities, etc. closed?)
        if not name:
            print("No project name given.")
            return
        elif name not in [p.name for p in bw.projects]:
            print("Project does not exist: {}".format(name))
            return
        if name != bw.projects.current or reload:
            bw.projects.set_current(name)
            #signals.project_selected.emit()
            print("Loaded project:", name)
#
#     def get_new_project_name_dialog(self):
#         name, ok = QtWidgets.QInputDialog.getText(
#             None,
#             "Create new project",
#             "Name of new project:" + " " * 25
#         )
#         return name if ok else None
#
#     def new_project(self, name=None):
#         name = name or self.get_new_project_name_dialog()
#         if name and name not in bw.projects:
#             bw.projects.set_current(name)
#             self.change_project(name, reload=True)
#             signals.projects_changed.emit()
#         elif name in bw.projects:
#             QtWidgets.QMessageBox.information(None,
#                                               "Not possible.",
#                                               "A project with this name already exists.")
#
#     def copy_project(self):
#         name, ok = QtWidgets.QInputDialog.getText(
#             None,
#             "Copy current project",
#             "Copy current project ({}) to new name:".format(bw.projects.current) + " " * 10
#         )
#         if ok and name:
#             if name not in bw.projects:
#                 bw.projects.copy_project(name, switch=True)
#                 self.change_project(name)
#                 signals.projects_changed.emit()
#             else:
#                 QtWidgets.QMessageBox.information(None,
#                                                   "Not possible.",
#                                                   "A project with this name already exists.")
#
#     def confirm_project_deletion_dialog(self):
#         buttonReply = QtWidgets.QMessageBox.question(
#             None,
#             'Confirm project deletion',
#             ("Are you sure you want to delete project '{}'? It has {} databases" +
#              " and {} LCI methods").format(
#                 bw.projects.current,
#                 len(bw.databases),
#                 len(bw.methods)
#             )
#         )
#         return buttonReply
#
#     def delete_project(self):
#         if len(bw.projects) == 1:
#             QtWidgets.QMessageBox.information(None,
#                                               "Not possible",
#                                               "Can't delete last project.")
#             return
#         buttonReply = self.confirm_project_deletion_dialog()
#         # truely deleting does not work due to a permission error on Windows
#         # delete_dir = QtWidgets.QMessageBox.question(None,
#         #     'Confirm project deletion',
#         #     "Do you also want to delete the files on your hard drive?")
#         if buttonReply == QtWidgets.QMessageBox.Yes:
#             bw.projects.delete_project(bw.projects.current, delete_dir=False)
#             self.change_project(futura_settings.startup_project, reload=True)
#             # if delete_dir:  # also purging does not work (PermissionError)
#             #     bw.projects.purge_deleted_directories()
#             signals.projects_changed.emit()
#
# # DATABASE
#     def ensure_sqlite_indices(self):
#         """
#         - fix for https://github.com/LCA-ActivityBrowser/activity-browser/issues/189
#         - also see bw2data issue: https://bitbucket.org/cmutel/brightway2-data/issues/60/massive-sqlite-query-performance-decrease
#         @LegacyCode?
#         """
#         if bw.databases and not sqlite3_lci_db._database.get_indexes('activitydataset'):
#             print('creating missing sqlite indices')
#             bw.Database(list(bw.databases)[-1])._add_indices()
#
#     def install_default_data(self):
#         self.default_biosphere_dialog = DefaultBiosphereDialog()
#         project_settings.add_db("biosphere3")
#
#     def add_database(self):
#         name, ok = QtWidgets.QInputDialog.getText(
#             None,
#             "Create new database",
#             "Name of new database:" + " " * 25
#         )
#
#         if ok and name:
#             if name not in bw.databases:
#                 bw.Database(name).register()
#                 project_settings.add_db(name)
#                 signals.databases_changed.emit()
#                 signals.database_selected.emit(name)
#             else:
#                 QtWidgets.QMessageBox.information(None,
#                                                   "Not possible",
#                                                   "A database with this name already exists.")
#
#     def copy_database(self, name):
#         new_name, ok = QtWidgets.QInputDialog.getText(
#             None,
#             "Copy {}".format(name),
#             "Name of new database:" + " " * 25)
#         if ok and new_name:
#             if new_name not in bw.databases:
#                 self.copydb_dialog = CopyDatabaseDialog(name, new_name)
#                 project_settings.add_db(new_name)
#             else:
#                 QtWidgets.QMessageBox.information(None,
#                                                   "Not possible",
#                                                   'Database <b>{}</b> already exists!'.format(new_name))
#
#     def delete_database(self, name):
#         ok = QtWidgets.QMessageBox.question(
#             None,
#             "Delete database?",
#             ("Are you sure you want to delete database '{}'? It has {} activity datasets").format(
#                 name, len(bw.Database(name)))
#         )
#         if ok == QtWidgets.QMessageBox.Yes:
#             project_settings.remove_db(name)
#             del bw.databases[name]
#             self.change_project(bw.projects.current, reload=True)
#
# # CALCULATION SETUP
#     def new_calculation_setup(self):
#         name, ok = QtWidgets.QInputDialog.getText(
#             None,
#             "Create new calculation setup",
#             "Name of new calculation setup:" + " " * 10
#         )
#         if ok and name:
#             if name not in bw.calculation_setups.keys():
#                 bw.calculation_setups[name] = {'inv': [], 'ia': []}
#                 signals.calculation_setup_selected.emit(name)
#                 print("New calculation setup: {}".format(name))
#             else:
#                 QtWidgets.QMessageBox.information(None,
#                                                   "Not possible",
#                                                   "A calculation setup with this name already exists.")
#
#     def delete_calculation_setup(self, name):
#         del bw.calculation_setups[name]
#         signals.set_default_calculation_setup.emit()
#         print("Deleted calculation setup: {}".format(name))
#
#     def rename_calculation_setup(self, current):
#         new_name, ok = QtWidgets.QInputDialog.getText(
#             None,
#             "Rename '{}'".format(current),
#             "New name of this calculation setup:" + " " * 10
#         )
#         if ok and new_name:
#             bw.calculation_setups[new_name] = bw.calculation_setups[current].copy()
#             # print("Current setups:", list(bw.calculation_setups.keys()))
#             del bw.calculation_setups[current]
#             # print("After deletion of {}:".format(current), list(bw.calculation_setups.keys()))
#             signals.calculation_setup_selected.emit(new_name)
#             print("Renamed calculation setup from {} to {}".format(current, new_name))
#
# # ACTIVITY
#     def new_activity(self, database_name):
#         # TODO: let user define product
#         name, ok = QtWidgets.QInputDialog.getText(
#             None,
#             "Create new technosphere activity",
#             "Please specify an activity name:" + " " * 10,
#         )
#         if ok and name:
#             new_act = bw.Database(database_name).new_activity(
#                 code=uuid.uuid4().hex,
#                 name=name,
#                 unit="unit",
#                 type="process",
#             )
#             new_act.save()
#             production_exchange = new_act.new_exchange(amount=1, type="production")
#             production_exchange.input = new_act
#             production_exchange.save()
#             signals.open_activity_tab.emit(new_act.key)
#             signals.metadata_changed.emit(new_act.key)
#             signals.database_changed.emit(database_name)
#             signals.databases_changed.emit()
#
#     def delete_activity(self, key):
#         act = bw.get_activity(key)
#         nu = len(act.upstream())
#         if nu:
#             text = "activities consume" if nu > 1 else "activity consumes"
#             QtWidgets.QMessageBox.information(
#                 None,
#                 "Not possible.",
#                 """Can't delete {}. {} upstream {} its reference product.
#                 Upstream exchanges must be modified or deleted.""".format(act, nu, text)
#             )
#         else:
#             act.delete()
#             signals.metadata_changed.emit(act.key)
#             signals.database_changed.emit(act['database'])
#             signals.databases_changed.emit()
#             signals.calculation_setup_changed.emit()
#
#     @staticmethod
#     def generate_copy_code(key: tuple) -> str:
#         db, code = key
#         metadata = AB_metadata.get_database_metadata(db)
#         if '_copy' in code:
#             code = code.split('_copy')[0]
#         copies = metadata["key"].apply(
#             lambda x: x[1] if code in x[1] and "_copy" in x[1] else None
#         ).dropna().to_list() if not metadata.empty else []
#         if not copies:
#             return "{}_copy1".format(code)
#         n = max((int(c.split('_copy')[1]) for c in copies))
#         return "{}_copy{}}".format(code, n+1)
#
#     def duplicate_activity(self, key):
#         """duplicates the selected activity in the same db, with a new BW code
#         for creating a copy in a different db, use copy_to_db"""
#         # todo: add "copy of" (or similar) to name of activity for easy identification in new db
#         # todo: some interface feedback so user knows the copy has succeeded
#         act = bw.get_activity(key)
#         new_code = self.generate_copy_code(key)
#         new_act = act.copy(new_code)
#         # Update production exchanges
#         for exc in new_act.production():
#             if exc.input.key == key:
#                 exc.input = new_act
#                 exc.save()
#         # Update 'products'
#         for product in new_act.get('products', []):
#             if product.get('input') == key:
#                 product['input'] = new_act.key
#         new_act.save()
#         signals.metadata_changed.emit(new_act.key)
#         signals.database_changed.emit(act['database'])
#         signals.databases_changed.emit()
#         signals.open_activity_tab.emit(new_act.key)
#
#     def show_duplicate_to_db_interface(self, activity_key):
#         origin_db = activity_key[0]
#         activity = bw.get_activity(activity_key)
#
#         available_target_dbs = list(project_settings.get_editable_databases())
#
#         if origin_db in available_target_dbs:
#             available_target_dbs.remove(origin_db)
#
#         if not available_target_dbs:
#             QtWidgets.QMessageBox.information(
#                 None,
#                 "No target database",
#                 "No valid target databases available. Create a new database or set one to writable (not read-only)."
#             )
#         else:
#             target_db, ok = QtWidgets.QInputDialog.getItem(
#                 None,
#                 "Copy activity to database",
#                 "Target database:",
#                 available_target_dbs,
#                 0,
#                 False
#             )
#             if ok:
#                 self.duplicate_activity_to_db(target_db, activity)
#
#     def duplicate_activity_to_db(self, target_db, activity):
#         new_code = self.generate_copy_code((target_db, activity['code']))
#         new_act_key = (target_db, new_code)
#         activity.copy(code=new_code, database=target_db)
#         # only process database immediately if small
#         if len(bw.Database(target_db)) < 50:
#             bw.databases.clean()
#
#         signals.metadata_changed.emit(new_act_key)
#         signals.database_changed.emit(target_db)
#         signals.open_activity_tab.emit(new_act_key)
#         signals.databases_changed.emit()
#
#     def modify_activity(self, key, field, value):
#         activity = bw.get_activity(key)
#         activity[field] = value
#         activity.save()
#         signals.metadata_changed.emit(key)
#         signals.database_changed.emit(key[0])
#
#     # def modify_exchanges_output(self, exchanges, key):
#     #     db_changed = {key[0]}
#     #     for exc in exchanges:
#     #         db_changed.add(exc['output'][0])
#     #         if exc['type'] == 'production':
#     #             new_exc = Exchange()
#     #             new_exc._data = copy.deepcopy(exc._data)
#     #             new_exc['type'] = 'technosphere'
#     #             new_exc['output'] = key
#     #             new_exc.save()
#     #         else:
#     #             exc['output'] = key
#     #             exc.save()
#     #     for db in db_changed:
#     #         signals.database_changed.emit(db)
#
#     def add_exchanges(self, from_keys: Iterable, to_key: tuple) -> None:
#         activity = bw.get_activity(to_key)
#         for key in from_keys:
#             technosphere_db = bc.is_technosphere_db(key[0])
#             exc = activity.new_exchange(input=key, amount=1)
#             if key == to_key:
#                 exc['type'] = 'production'
#             elif technosphere_db is True:
#                 exc['type'] = 'technosphere'
#             elif technosphere_db is False:
#                 exc['type'] = 'biosphere'
#             else:
#                 exc['type'] = 'unknown'
#             exc.save()
#         signals.metadata_changed.emit(to_key)
#         signals.database_changed.emit(to_key[0])
#
#
#     def delete_exchanges(self, exchanges):
#         db_changed = set()
#         for exc in exchanges:
#             db_changed.add(exc['output'][0])
#             exc._document.delete_instance()
#         for db in db_changed:
#             # signals.metadata_changed.emit(to_key)
#             signals.database_changed.emit(db)
#
#     def modify_exchange_amount(self, exchange, value):
#         exchange['amount'] = value
#         exchange.save()
#
#         signals.database_changed.emit(exchange['output'][0])
#
#     @staticmethod
#     @Slot(object, str, object)
#     def modify_exchange(exchange, field, value):
#         # The formula field needs special handling.
#         if field == "formula":
#             if field in exchange and (value == "" or value is None):
#                 # Remove formula entirely.
#                 del exchange[field]
#                 if "original_amount" in exchange:
#                     # Restore the original amount, if possible
#                     exchange["amount"] = exchange["original_amount"]
#                     del exchange["original_amount"]
#             if value:
#                 # At least set the formula, possibly also store the amount
#                 if field not in exchange:
#                     exchange["original_amount"] = exchange["amount"]
#                 exchange[field] = value
#         else:
#             exchange[field] = value
#         exchange.save()
#         if field == "formula" and value:
#             # If a formula was set or changed, recalculate exchanges
#             signals.exchange_formula_changed.emit(exchange["output"])
#         signals.database_changed.emit(exchange['output'][0])
    def run_dialog(self, parent=None):
        print('run_dialog')
        rundialog = RunDialog(parent)
        px = parent.geometry().x()
        py = parent.geometry().y()
        pw = parent.width()
        ph = parent.height()
        print(px,py, pw, ph)
        w = 150
        h = 150
        x = round(px + pw/2 - w/2)
        y = round(py + ph/2 - h/2)
        rundialog.setGeometry(x, y, w, h)
        rundialog.exec_()


    def load_tree_data(self):

        projectDialog = ProjectDialog()
        #projectDialog.show()
        projectDialog.exec_()
        print('signal recieved')
        print('loading tree')
        #loading = LoadingDialog(self)
        #loading.setModal(False)
        #loading.show()
        print('loading showing')
        # might need to send this to a different thread maybe singleshot QTimer?
        #from futura.constants import TREE_DATA
        print('import finished')
        #self.tree_model = TreeModel(TREE_DATA)
        #print('treemodel created')
        #self.ui.treeView.setModel(self.tree_model)
        print('tree loaded')
        #loading.close()

    def project_selected(self):
        print('selected a project')
        print(sorted([d for d in bw.databases]))




# -*- coding: utf-8 -*-
import json
import os
import shutil
from typing import Optional

import appdirs
import brightway2 as bw

from futura_ui.app.signals import signals
from .. import PACKAGE_DIRECTORY


class BaseSettings(object):
    """ Base Class for handling JSON settings files.
    """
    def __init__(self, directory: str, filename: str = None):
        self.data_dir = directory
        self.filename = filename or "default_settings.json"
        self.settings_file = os.path.join(self.data_dir, self.filename)
        self.settings: Optional[dict] = None
        self.initialize_settings()

    @classmethod
    def get_default_settings(cls) -> dict:
        """ Returns dictionary containing the default settings for the file
        """
        raise NotImplementedError

    def restore_default_settings(self) -> None:
        """ Undo all user settings and return to original state.
        """
        self.settings = self.get_default_settings()
        self.write_settings()

    def initialize_settings(self) -> None:
        """ Attempt to find and read the settings_file, creates a default
        if not found
        """
        if os.path.isfile(self.settings_file):
            self.load_settings()
        else:
            self.settings = self.get_default_settings()
            self.write_settings()

    def load_settings(self) -> None:
        with open(self.settings_file, "r") as infile:
            self.settings = json.load(infile)

    def write_settings(self) -> None:
        with open(self.settings_file, "w") as outfile:
            json.dump(self.settings, outfile, indent=4, sort_keys=True)


class FuturaSettings(BaseSettings):
    """
    Interface to the json settings file. Will create a userdata directory via appdirs if not
    already present.
    """
    def __init__(self, filename: str):
        futura_dir = appdirs.AppDirs("Futura", "Futura")
        if not os.path.isdir(futura_dir.user_data_dir):
            os.makedirs(futura_dir.user_data_dir, exist_ok=True)

        super().__init__(futura_dir.user_data_dir, filename)

    @classmethod
    def get_default_settings(cls) -> dict:
        """ Using methods from the commontasks file to set default settings
        EDIT THIS TO REFLECT FUTURA SETTINGS
        """
        return {
            "custom_bw_dir": cls.get_default_directory(),
            "startup_project": cls.get_default_project_name(),
        }

    # @property
    # def custom_bw_dir(self) -> str:
    #     """ Returns the custom brightway directory, or the default
    #     """
    #     return self.settings.get("custom_bw_dir", self.get_default_directory())
    #
    # @custom_bw_dir.setter
    # def custom_bw_dir(self, directory: str) -> None:
    #     """ Sets the custom brightway directory to `directory`
    #     """
    #     self.settings.update({"custom_bw_dir": directory})

    # @property
    # def startup_project(self) -> str:
    #     """ Get the startup project from the settings, or the default
    #     """
    #     project = self.settings.get(
    #         "startup_project", self.get_default_project_name()
    #     )
    #     if project not in bw.projects:
    #         project = self.get_default_project_name()
    #     return project

    # @startup_project.setter
    # def startup_project(self, project: str) -> None:
    #     """ Sets the startup project to `project`
    #     """
    #     self.settings.update({"startup_project": project})

    @staticmethod
    def get_default_directory() -> str:
        """ Returns the default brightway application directory
        """
        return None # bw.projects._get_base_directories()[0]

    @staticmethod
    def get_default_project_name() -> Optional[str]:
        """ Returns the default project name.
        """
        return None
        # if "default" in bw.projects:
        #     return "default"
        # elif len(bw.projects):
        #     return next(iter(bw.projects)).name
        # else:
        #     return None


# class ProjectSettings(BaseSettings):
#     """
#     Handles user settings which are specific to projects. Created initially to handle read-only/writable database status
#     Code based on ABSettings class, if more different types of settings are needed, could inherit from a base class
#
#     structure: singleton, loaded dependent on which project is selected.
#         Persisted on disc, Stored in the BW2 projects data folder for each project
#         a dictionary1 of dictionaries2
#         Dictionary1 keys are settings names (currently just 'read-only-databases'), values are dictionary2s
#         Dictionary2 keys are database names, values are bools
#
#     For now, decided to not include saving writable-activities to settings.
#     As activities are identified by tuples, and saving them to json requires extra code
#     https://stackoverflow.com/questions/15721363/preserve-python-tuples-with-json
#     This is currently not worth the effort but could be returned to later
#
#     """
#     def __init__(self, filename: str):
#         # on selection of a project (signal?), find the settings file for that project if it exists
#         # it can be a custom location, based on ABsettings. So check that, and if not, use default?
#         # once found, load the settings or just an empty dict.
#         self.connect_signals()
#         super().__init__(bw.projects.dir, filename)
#
#         # https://github.com/LCA-ActivityBrowser/activity-browser/issues/235
#         # Fix empty settings file and populate with currently active databases
#         if "read-only-databases" not in self.settings:
#             self.settings.update(self.process_brightway_databases())
#             self.write_settings()
#
#     def connect_signals(self):
#         """ Reload the project settings whenever a project switch occurs.
#         """
#         signals.project_selected.connect(self.reset_for_project_selection)
#         signals.delete_project.connect(self.reset_for_project_selection)
#
#     @classmethod
#     def get_default_settings(cls) -> dict:
#         """ Return default empty settings dictionary.
#         """
#         return cls.process_brightway_databases()
#
#     @staticmethod
#     def process_brightway_databases() -> dict:
#         """ Process brightway database list and return new settings dictionary.
#
#         NOTE: This ignores the existing database read-only settings.
#         """
#         return {
#             "read-only-databases": {name: True for name in bw.databases.list}
#         }
#
#     def reset_for_project_selection(self) -> None:
#         """ On switching project, attempt to read the settings for the new
#         project.
#         """
#         print("Reset project settings directory to:", bw.projects.dir)
#         self.settings_file = os.path.join(bw.projects.dir, self.filename)
#         self.initialize_settings()
#
#     def add_db(self, db_name: str, read_only: bool = True) -> None:
#         """ Store new databases and relevant settings here when created/imported
#         """
#         self.settings["read-only-databases"].setdefault(db_name, read_only)
#         self.write_settings()
#
#     def modify_db(self, db_name: str, read_only: bool) -> None:
#         """ Update write-rules for the given database
#         """
#         self.settings["read-only-databases"].update({db_name: read_only})
#         self.write_settings()
#
#     def remove_db(self, db_name: str) -> None:
#         """ When a database is deleted from a project, the settings are also deleted.
#         """
#         self.settings["read-only-databases"].pop(db_name, None)
#         self.write_settings()
#
#     def db_is_readonly(self, db_name: str) -> bool:
#         """ Check if given database is read-only, defaults to yes.
#         """
#         return self.settings["read-only-databases"].get(db_name, True)
#
#     def get_editable_databases(self):
#         """ Return list of database names where read-only is false
#
#         NOTE: discards the biosphere3 database based on name.
#         """
#         iterator = self.settings.get("read-only-databases", {}).items()
#         return (name for name, ro in iterator if not ro and name != "biosphere3")


futura_settings = FuturaSettings("FuturaSettings.json")
# project_settings = ProjectSettings("AB_project_settings.json")



import os
import appdirs
import glob

class FuturaStorage():
    def __init__(self):
        self.futura_dir = appdirs.user_data_dir(
            appname='Futura',
            appauthor='Futura'
        )
        if not os.path.isdir(self.futura_dir):
            os.makedirs(self.futura_dir)

        # Databases
        self.data_dir = os.path.join(self.futura_dir, 'databases')
        if not os.path.isdir(self.data_dir):
            os.mkdir(self.data_dir)

        # recipe files
        self.recipe_dir = os.path.join(self.futura_dir, 'recipes')
        if not os.path.isdir(self.recipe_dir):
            os.mkdir(self.recipe_dir)

        # config
        # self.config_file = os.path.join(self.futura_dir, 'lcopt_config.yml')
        # if not os.path.exists(self.config_file):
        #     self.write_default_config()

        # self.config = self.load_config()

    @property
    def databases(self):
        databases = glob.glob(os.path.join(self.data_dir), '*.fdb')
        return databases

storage = FuturaStorage()


# from pathlib import Path
# from futura.sectors import *
# from futura.wrappers import FuturaDatabase
import os

ASSET_PATH = os.path.join(os.path.dirname(__file__), 'assets')
CCS_FILE = 'lci-Carma-CCS-base-GLO2.xlsx'
FULL_CCS_FILE = os.path.join(ASSET_PATH, CCS_FILE)

# ELECTRICITY_TREE = [build_tree(bw_glo_elec, filters=electricity_markets)]
# CEMENT_TREE = [build_tree(bw_cement_europe, filters=cement_filter),
#                build_tree(bw_cement_US, filters=cement_filter),]
#
#
# TREE_DATA = {'Electricity': ELECTRICITY_TREE,
#              'Cement': CEMENT_TREE,
#              'Aluminium': {},
#              'Steel': {},
#              'Lithium': {},
#              'Glass': {}
#              }




from .regionalisation import *
from functools import partial
from . import log, warn
from .default_filters import *
from tqdm import tqdm
from .constants import FULL_CCS_FILE
from.utils import create_filter_from_description


def add_technology_to_database(database, technology_file, funcs):

    assert type(database) == FuturaDatabase, "database needs to be a futura FuturaDatabase"

    database.extract_excel_data(technology_file)

    for f in funcs:
        database = f(database)


def fix_ch_only_processes(database):

    all_carma_filter = [w.equals('database', 'Carma CCS')]
    all_carma_filter += [w.equals('location', 'GLO')]

    all_carma_list = list(w.get_many(database.db, *all_carma_filter))

    ch_list = []
    for l in all_carma_list:
        for e in l['exchanges']:
            if 'location' in e.keys():
                if e['location'] == 'CH':
                    ch_list.append(e)

    ch_names_set = set([x['name'] for x in ch_list])
    ch_only_list = list(ch_names_set)

    for x in ch_only_list:
        found = w.get_many(database.db, *[w.contains('name', x)])
        found_list = list(found)
        location_list = [l['location'] for l in found_list]

        if all(["GLO" not in location_list, "RoW" not in location_list]):
            ch_only_process = [c for c in found_list if c['location'] == 'CH'][0]
            create_regional_activities(ch_only_process, ['GLO'], database.db)
            log("Adding {} [{}]".format(ch_only_process['name'], ch_only_process['location']))

    return database


def regionalise_multiple_processes(database, locations, base_activity_filter, progress_message=None):

    if not callable(base_activity_filter[0]):
        print('Creating base_activity_filter from description')
        base_activity_filter = create_filter_from_description(base_activity_filter)

    base_activities = list(w.get_many(database.db, *base_activity_filter))

    code_list = []
    if not progress_message:
        progress_message = 'specified activities'

    print("Creating regionalised versions of {}...".format(progress_message))
    for a in tqdm(base_activities):
        code_list.extend(create_regional_activities(a, locations, database.db, relink_now=False))

    print("Relinking regionalised versions of {}...".format(progress_message))
    for x in tqdm(code_list):
        ds = w.get_one(database.db, *[w.equals('code', x[1])])

        if ds['name'].startswith('market'):
            exclusive = False
        else:
            exclusive = True

        if 'input' in w.reference_product(ds).keys():
            del w.reference_product(ds)['input']

        w.relink_technosphere_exchanges(ds,
                                        database.db,
                                        exclusive=exclusive,
                                        drop_invalid=False,
                                        biggest_first=False,
                                        contained=False,
                                        exclude=['UCTE'])
    return database


def regionalise_based_on_filters(database, location_filter, base_activity_filter, progress_message=None):

    location_list = list(w.get_many(database.db, *location_filter))
    if len(set([x['name'] for x in location_list])) != 1:
        warn("The location filter returns more than one process")

    locations = list(set([x['location'] for x in location_list]))

    return regionalise_multiple_processes(database, locations, base_activity_filter, progress_message)


add_hard_coal_ccs = partial(regionalise_based_on_filters,
                            location_filter=coal_location_filter,
                            base_activity_filter=hard_coal_ccs_filter,
                            progress_message='hard coal CCS activities')

add_lignite_ccs = partial(regionalise_based_on_filters,
                          location_filter=lignite_location_filter,
                          base_activity_filter=lignite_ccs_filter,
                          progress_message='lignite CCS activities')

add_natural_gas_ccs = partial(regionalise_based_on_filters,
                              location_filter=natural_gas_location_filter,
                              base_activity_filter=natural_gas_ccs_filter,
                              progress_message='natural gas CCS activities')

add_wood_ccs = partial(regionalise_based_on_filters,
                       location_filter=wood_location_filter,
                       base_activity_filter=wood_ccs_filter,
                       progress_message='wood CCS activities')

add_default_CCS_processes = partial(add_technology_to_database,
                                    technology_file=FULL_CCS_FILE,
                                    funcs=[
                                     fix_ch_only_processes,
                                     add_hard_coal_ccs,
                                     add_lignite_ccs,
                                     add_natural_gas_ccs,
                                     add_wood_ccs
                                    ])



import logging
# import wurst
import wurst
import futura.wurst_monkeypatch as wmp
#do monkeypatching here

wurst.relink_technosphere_exchanges = wmp.relink_technosphere_exchanges
wurst.transformations.relink_technosphere_exchanges = wmp.relink_technosphere_exchanges
wurst.transformations.geo.relink_technosphere_exchanges = wmp.relink_technosphere_exchanges
wurst.transformations.geo.allocate_inputs = wmp.allocate_inputs

#create w alias for wurst
w = wurst

# set up logging to file - see previous section for more details
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M',
                    filename=r'C:\temp\futura.log',
                    filemode='w')
# define a Handler which writes INFO messages or higher to the sys.stderr
console = logging.StreamHandler()
console.setLevel(logging.WARNING)
# set a format which is simpler for console use
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
# tell the handler to use this format
console.setFormatter(formatter)
# add the handler to the root logger
logging.getLogger('').addHandler(console)

logger1 = logging.getLogger('futura')


def log(MESSAGE, *args, **kwargs):
    logger1.info(MESSAGE, *args, **kwargs)


def warn(MESSAGE, *args, **kwargs):
    logger1.warning(MESSAGE, *args, **kwargs)

warn('Futura is using a monkeypatched version of Wurst')



# -*- coding: utf-8 -*-
import os

from PySide2.QtGui import QIcon

from futura_ui import PACKAGE_DIRECTORY


def create_path(folder: str, filename: str) -> str:
    """ Builds a path to the image file.
    """
    return os.path.join(PACKAGE_DIRECTORY, "icons", folder, filename)


# CURRENTLY UNUSED ICONS

    # Modular LCA (keep until this is reintegrated)
    # add_db = create_path('metaprocess', 'add_database.png')
    # close_db = create_path('metaprocess', 'close_database.png')
    # cut = create_path('metaprocess', 'cut.png')
    # duplicate = create_path('metaprocess', 'duplicate.png')
    # graph_lmp = create_path('metaprocess', 'graph_linkedmetaprocess.png')
    # graph_mp = create_path('metaprocess', 'graph_metaprocess.png')
    # load_db = create_path('metaprocess', 'open_database.png')
    # metaprocess = create_path('metaprocess', 'metaprocess.png')
    # new = create_path('metaprocess', 'new_metaprocess.png')
    # save_db = create_path('metaprocess', 'save_database.png')
    # save_mp = create_path('metaprocess', 'save_metaprocess.png')

    # key = create_path('main', 'key.png')
    # search = create_path('main', 'search.png')
    # switch = create_path('main', 'switch-state.png')


class Icons(object):
    # Icons from href="https://www.flaticon.com/

    # MAIN
    futura = create_path('main', 'main_icon.png')

    # arrows
    # right = create_path('main', 'right.png')
    # left = create_path('main', 'left.png')
    # forward = create_path('main', 'forward.png')
    # backward = create_path('main', 'backward.png')

    # Simple actions
    # delete = create_path('context', 'delete.png')
    # copy = create_path('context', 'copy.png')
    # add = create_path('context', 'add.png')
    # edit = create_path('main', 'edit.png')
    # calculate = create_path('main', 'calculate.png')
    # question = create_path('context', 'question.png')

    # database
    # import_db = create_path('main', 'import_database.png')
    # duplicate_database = create_path('main', 'duplicate_database.png')

    # activity
    # duplicate_activity = create_path('main', 'duplicate_activity.png')
    # duplicate_to_other_database = create_path('main', 'import_database.png')

    # windows
    # graph_explorer = create_path('main', 'graph_explorer.png')
    # debug = create_path('main', 'ladybird.png')
    # issue = create_path('main', 'idea.png')
    # settings = create_path('main', 'settings.png')
    # history = create_path('main', 'history.png')
    # welcome = create_path('main', 'welcome.png')
    # main_window = create_path('main', 'home.png')


class QIcons(Icons):
    """ Using the Icons class, returns the same attributes, but as QIcon type
    """
    def __getattribute__(self, item):
        return QIcon(Icons.__getattribute__(self, item))


icons = Icons()
qicons = QIcons()



import yaml
import jinja2

from futura.markets import FuturaMarket
from futura.utils import create_filter_from_description
from .wrappers import FuturaDatabase
from . import technology
from .constants import ASSET_PATH
# import wurst as w
from . import w
from .regionalisation import create_regional_activities

import os.path

class FuturaLoader:

    def __init__(self, recipe_filepath, autocreate=True):

        self.recipe = self.load_recipe(recipe_filepath)
        self.database = None

        if autocreate:
            self.run()

    @staticmethod
    def load_recipe(filename):

        with open(filename, "r") as f:
            template = jinja2.Template(f.read())

        t_data = template.render()
        data = yaml.load(t_data, yaml.Loader)

        return data

    def parse_load_section(self):

        assert 'load' in self.recipe.keys(), "No load section to parse"

        possible_functions = [
            'extract_bw2_database',
            'extract_excel_data'
        ]

        wd = FuturaDatabase()

        for f in self.recipe['load']:
            for k, v in f.items():

                assert k in possible_functions, "{} is not a valid load function".format(k)

                if k == 'extract_bw2_database':
                    wd.extract_bw2_database(**v)

                elif k == 'extract_excel_data':
                    wd.extract_excel_data(**v)

        return wd

    def parse_technology_section(self):

        assert 'technology' in self.recipe.keys(), "No technology section to parse"

        possible_functions = [
            'add_technology_to_database',
            'add_default_CCS_processes',
        ]

        for f in self.recipe['technology']:
            for k, v in f.items():

                assert k in possible_functions, "{} is not a valid technology function".format(k)

                if k == 'add_technology_to_database':

                    possible_subfunctions = [
                        'create_regional_activities'
                    ]
                    actual_functions = []
                    for func in v['tasks']:
                        if func['function'] in possible_subfunctions:
                            if func['function'] == 'create_regional_activities':
                                this_item = {'function': create_regional_activities,
                                             'args': func.get('args', []),
                                             'kwargs': func.get('kwargs', {})}

                                actual_functions.append(this_item)

                        elif func['function'] in dir(technology):
                            args = [self.database]
                            args.extend(func.get('args', []))
                            this_item = {'function': getattr(technology, func['function']),
                                         'args': args,
                                         'kwargs': func.get('kwargs', {})}
                            actual_functions.append(this_item)

                    if "__ASSET_PATH__/" in v['technology_file']:
                        technology_file = v['technology_file'].replace("__ASSET_PATH__/", "")
                        technology_path = os.path.join(ASSET_PATH, technology_file)
                    else:
                        technology_path = v['technology_file']

                    self.database.extract_excel_data(technology_path)

                    for item in actual_functions:
                        _ = item['function'](*item['args'], **item['kwargs'])
                    # technology.add_technology_to_database(self.database, technology_path, actual_functions)

                elif k == 'add_default_CCS_processes':
                    technology.add_default_CCS_processes(self.database)

    def parse_market_section(self):

        assert 'markets' in self.recipe.keys(), "No markets section to parse"

        possible_functions = [
            'alter_market',
        ]

        for f in self.recipe['markets']:
            for k, v in f.items():

                assert k in possible_functions, "{} is not a valid market function".format(k)

                if k == 'alter_market':
                    market_filter = create_filter_from_description(v['market_filter'])
                    market = w.get_one(self.database.db, *market_filter)
                    fm = FuturaMarket(market, self.database)

                    for task in v['tasks']:

                        message = "Applying {}".format(task['function'])

                        if task['args']:
                            message += " with arguments {}".format(", ".join([str(x) for x in task['args']]))

                        if 'kwargs' in task.keys():
                            kwargs = task['kwargs']
                            message += " and keyword arguments {}".format(
                                ", ".join(["{}:{}".format(k, v) for k, v in kwargs.items()])
                            )
                        else:
                            kwargs = {}

                        print(message)
                        func = getattr(fm, task['function'])
                        func(*task['args'], **kwargs)

    def run(self):

        self.database = self.parse_load_section()

        if 'technology' in self.recipe.keys():
            self.parse_technology_section()

        if 'markets' in self.recipe.keys():
            self.parse_market_section()

    def write_database(self, project=None, database=None, overwrite=True):
        assert isinstance(self.database, FuturaDatabase)
        assert 'metadata' in self.recipe.keys()
        if not project:
            assert 'base_project' in self.recipe['metadata'].keys()
            project = self.recipe['metadata']['base_project']
        if not database:
            assert 'output_database' in self.recipe['metadata'].keys()
            database = self.recipe['metadata']['output_database']

        self.database.write_database(project, database, overwrite)
